@@ -1,5 +1,5 @@
 /* GNU gettext - internationalization aids
-   Copyright (C) 1995-1999, 2000-2003 Free Software Foundation, Inc.
+   Copyright (C) 1995-1999, 2000-2006 Free Software Foundation, Inc.
 
    This file was written by Peter Miller <millerp@canb.auug.org.au>.
    Multibyte character handling by Bruno Haible <haible@clisp.cons.org>.
@@ -16,7 +16,7 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software Foundation,
-   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
 
 
 #ifdef HAVE_CONFIG_H
@@ -46,7 +46,11 @@
 #include "exit.h"
 #include "error.h"
 #include "error-progname.h"
+#include "xvasprintf.h"
+#include "po-error.h"
+#include "po-xerror.h"
 #include "pos.h"
+#include "message.h"
 #include "str-list.h"
 #include "po-gram-gen2.h"
 
@@ -70,13 +74,6 @@ int gram_pos_column;
 /* Error handling during the parsing of a PO file.
    These functions can access gram_pos and gram_pos_column.  */
 
-#if !(__STDC__ && \
-      ((defined __STDC_VERSION__ && __STDC_VERSION__ >= 199901L && !defined __DECC) \
-       || (defined __GNUC__ && __GNUC__ >= 2 && !defined __APPLE_CC__)))
-
-/* CAUTION: If you change this function, you must also make identical
-   changes to the macro of the same name in src/po-lex.h  */
-
 /* VARARGS1 */
 void
 po_gram_error (const char *fmt, ...)
@@ -88,24 +85,14 @@ po_gram_error (const char *fmt, ...)
   if (vasprintf (&buffer, fmt, ap) < 0)
     error (EXIT_FAILURE, 0, _("memory exhausted"));
   va_end (ap);
-  error_with_progname = false;
-  error (0, 0, "%s:%lu:%d: %s", gram_pos.file_name,
-	 (unsigned long) gram_pos.line_number, gram_pos_column + 1, buffer);
-  error_with_progname = true;
+  po_xerror (PO_SEVERITY_ERROR, NULL, gram_pos.file_name, gram_pos.line_number,
+	     gram_pos_column + 1, false, buffer);
   free (buffer);
 
-  /* Some messages need more than one line.  Continuation lines are
-     indicated by using "..." at the start of the string.  We don't
-     increment the error counter for these continuation lines.  */
-  if (*fmt == '.')
-    --error_message_count;
-  else if (error_message_count >= gram_max_allowed_errors)
-    error (EXIT_FAILURE, 0, _("too many errors, aborting"));
+  if (error_message_count >= gram_max_allowed_errors)
+    po_error (EXIT_FAILURE, 0, _("too many errors, aborting"));
 }
 
-/* CAUTION: If you change this function, you must also make identical
-   changes to the macro of the same name in src/po-lex.h  */
-
 /* VARARGS2 */
 void
 po_gram_error_at_line (const lex_pos_ty *pp, const char *fmt, ...)
@@ -117,23 +104,14 @@ po_gram_error_at_line (const lex_pos_ty *pp, const char *fmt, ...)
   if (vasprintf (&buffer, fmt, ap) < 0)
     error (EXIT_FAILURE, 0, _("memory exhausted"));
   va_end (ap);
-  error_with_progname = false;
-  error_at_line (0, 0, pp->file_name, pp->line_number, "%s", buffer);
-  error_with_progname = true;
+  po_xerror (PO_SEVERITY_ERROR, NULL, pp->file_name, pp->line_number,
+	     (size_t)(-1), false, buffer);
   free (buffer);
 
-  /* Some messages need more than one line, or more than one location.
-     Continuation lines are indicated by using "..." at the start of the
-     string.  We don't increment the error counter for these
-     continuation lines.  */
-  if (*fmt == '.')
-    --error_message_count;
-  else if (error_message_count >= gram_max_allowed_errors)
-    error (EXIT_FAILURE, 0, _("too many errors, aborting"));
+  if (error_message_count >= gram_max_allowed_errors)
+    po_error (EXIT_FAILURE, 0, _("too many errors, aborting"));
 }
 
-#endif
-
 
 /* The lowest level of PO file parsing converts bytes to multibyte characters.
    This is needed
@@ -310,7 +288,11 @@ mb_width (const mbchar_t mbc)
     {
       if (mbc->bytes == 1)
 	{
-	  if (mbc->buf[0] >= 0x00 && mbc->buf[0] <= 0x1F)
+	  if (
+#if CHAR_MIN < 0x00 /* to avoid gcc warning */
+	      mbc->buf[0] >= 0x00 &&
+#endif
+	      mbc->buf[0] <= 0x1F)
 	    {
 	      if (mbc->buf[0] == 0x09)
 		return 8 - (gram_pos_column & 7);
@@ -501,7 +483,13 @@ incomplete multibyte sequence at end of line"));
 		    }
 		}
 	      else
-		error (EXIT_FAILURE, errno, _("iconv failure"));
+		{
+		  const char *errno_description = strerror (errno);
+		  po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+			     xasprintf ("%s: %s",
+					_("iconv failure"),
+					errno_description));
+		}
 	    }
 	  else
 	    {
@@ -614,6 +602,7 @@ mbfile_ungetc (const mbchar_t mbc, mbfile_t mbf)
 static mbfile_t mbf;
 unsigned int gram_max_allowed_errors = 20;
 static bool po_lex_obsolete;
+static bool po_lex_previous;
 static bool pass_comments = false;
 bool pass_obsolete_entries = false;
 
@@ -632,6 +621,7 @@ lex_start (FILE *fp, const char *real_filename, const char *logical_filename)
   gram_pos_column = 0;
   signal_eilseq = true;
   po_lex_obsolete = false;
+  po_lex_previous = false;
   po_lex_charset_init ();
 }
 
@@ -645,6 +635,7 @@ lex_end ()
   gram_pos_column = 0;
   signal_eilseq = false;
   po_lex_obsolete = false;
+  po_lex_previous = false;
   po_lex_charset_close ();
 }
 
@@ -661,10 +652,14 @@ lex_getc (mbchar_t mbc)
       if (mb_iseof (mbc))
 	{
 	  if (ferror (mbf->fp))
+	   bomb:
 	    {
-	    bomb:
-	      error (EXIT_FAILURE, errno, _("error while reading \"%s\""),
-		     gram_pos.file_name);
+	      const char *errno_description = strerror (errno);
+	      po_xerror (PO_SEVERITY_FATAL_ERROR, NULL, NULL, 0, 0, false,
+			 xasprintf ("%s: %s",
+				    xasprintf (_("error while reading \"%s\""),
+					       gram_pos.file_name),
+				    errno_description));
 	    }
 	  break;
 	}
@@ -726,14 +721,29 @@ lex_ungetc (const mbchar_t mbc)
 static int
 keyword_p (const char *s)
 {
-  if (!strcmp (s, "domain"))
-    return DOMAIN;
-  if (!strcmp (s, "msgid"))
-    return MSGID;
-  if (!strcmp (s, "msgid_plural"))
-    return MSGID_PLURAL;
-  if (!strcmp (s, "msgstr"))
-    return MSGSTR;
+  if (!po_lex_previous)
+    {
+      if (!strcmp (s, "domain"))
+	return DOMAIN;
+      if (!strcmp (s, "msgid"))
+	return MSGID;
+      if (!strcmp (s, "msgid_plural"))
+	return MSGID_PLURAL;
+      if (!strcmp (s, "msgstr"))
+	return MSGSTR;
+      if (!strcmp (s, "msgctxt"))
+	return MSGCTXT;
+    }
+  else
+    {
+      /* Inside a "#|" context, the keywords have a different meaning.  */
+      if (!strcmp (s, "msgid"))
+	return PREV_MSGID;
+      if (!strcmp (s, "msgid_plural"))
+	return PREV_MSGID_PLURAL;
+      if (!strcmp (s, "msgctxt"))
+	return PREV_MSGCTXT;
+    }
   po_gram_error_at_line (&gram_pos, _("keyword \"%s\" unknown"), s);
   return NAME;
 }
@@ -872,6 +882,7 @@ po_gram_lex ()
 	  {
 	  case '\n':
 	    po_lex_obsolete = false;
+	    po_lex_previous = false;
 	    /* Ignore whitespace, not relevant for the grammar.  */
 	    break;
 
@@ -892,6 +903,24 @@ po_gram_lex ()
 		 characters are expected to be well formed.  */
 	      {
 		po_lex_obsolete = true;
+		/* A pseudo-comment beginning with #~| denotes a previous
+		   untranslated string in an obsolete entry.  This does not
+		   make much sense semantically, and is implemented here
+		   for completeness only.  */
+		lex_getc (mbc);
+		if (mb_iseq (mbc, '|'))
+		  po_lex_previous = true;
+		else
+		  lex_ungetc (mbc);
+		break;
+	      }
+	    if (mb_iseq (mbc, '|'))
+	      /* A pseudo-comment beginning with #| is found.  This is
+		 the previous untranslated string.  We discard the "#|"
+		 prefix, but change the keywords and string returns
+		 accordingly.  */
+	      {
+		po_lex_previous = true;
 		break;
 	      }
 
@@ -930,7 +959,7 @@ po_gram_lex ()
 	      {
 		/* We do this in separate loop because collecting large
 		   comments while they get not passed to the upper layers
-		   is not very effective.  */
+		   is not very efficient.  */
 		while (!mb_iseof (mbc) && !mb_iseq (mbc, '\n'))
 		  lex_getc (mbc);
 		po_lex_obsolete = false;
@@ -975,11 +1004,17 @@ po_gram_lex ()
 	      }
 	    buf[bufpos] = '\0';
 
+	    /* Strings cannot contain the msgctxt separator, because it cannot
+	       be faithfully represented in the msgid of a .mo file.  */
+	    if (strchr (buf, MSGCTXT_SEPARATOR) != NULL)
+	      po_gram_error_at_line (&gram_pos,
+				     _("context separator <EOT> within string"));
+
 	    /* FIXME: Treatment of embedded \000 chars is incorrect.  */
 	    po_gram_lval.string.string = xstrdup (buf);
 	    po_gram_lval.string.pos = gram_pos;
 	    po_gram_lval.string.obsolete = po_lex_obsolete;
-	    return STRING;
+	    return (po_lex_previous ? PREV_STRING : STRING);
 
 	  case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
 	  case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
